/*
 * program_commons.c
 *
 * Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "program_commons.h"

static char *read_stdin (void);

TarotGame *
tarot_program_read_game (void)
{
  TarotGame *ret = tarot_game_alloc ();
  char *input = read_stdin ();
  if (tarot_game_load_from_xml (ret, input) == TAROT_GAME_OK)
    {
      free (input);
      return ret;
    }
  fprintf (stderr, _ ("Error: invalid game.\n"));
  abort ();
}

void
tarot_program_setup (void)
{
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  tarot_set_datadir (DATADIR);
  if (tarot_init (LOCALEDIR) != 0)
    {
      fprintf (stderr, _ ("ERROR: could not initialize libtarot.\n"));
    }
}

static size_t parse_unit (const char *str, char **end);

size_t
tarot_strtosz (const char *str, char **end)
{
  char *end_round;
  char *end_flt;
  size_t as_round = strtoul (str, &end_round, 0);
  double as_flt = strtod (str, &end_flt);
  size_t unit = 1;
  if (end_round >= end_flt)
    {
      /* Keep as_round, because this is an integer */
      unit = parse_unit (end_round, end);
      return as_round * unit;
    }
  else
    {
      /* This is a decimal number */
      unit = parse_unit (end_flt, end);
      return (size_t) (as_flt * unit);
    }
}

struct buffer;
typedef struct buffer buffer;
static buffer *buffer_alloc (void);
static void buffer_free (buffer *buffer);
static void buffer_push_back (buffer *b, char c);
static char *buffer_data (buffer *b); /* NOT nul-terminated */

static char *
read_stdin (void)
{
  buffer *buffer = buffer_alloc ();
  int c;
  char *ret;
  while ((c = getchar ()) != EOF)
    {
      unsigned int u = c;
      buffer_push_back (buffer, (unsigned char) u);
    }
  buffer_push_back (buffer, '\0');
  ret = strdup (buffer_data (buffer));
  buffer_free (buffer);
  return ret;
}

struct buffer
{
  size_t allocated;
  size_t used;
  char *data;
};

static buffer *
buffer_alloc (void)
{
  buffer *ret = xmalloc (sizeof (buffer));
  ret->allocated = 1;
  ret->used = 0;
  ret->data = xmalloc (1);
  return ret;
}

static void
buffer_free (buffer *buffer)
{
  free (buffer->data);
  free (buffer);
}

static void
buffer_push_back (buffer *b, char c)
{
  if (b->used >= b->allocated)
    {
      b->allocated *= 2;
      b->data = xrealloc (b->data, b->allocated);
    }
  assert (b->used < b->allocated);
  b->data[b->used++] = c;
}

static char *
buffer_data (buffer *b)
{
  return b->data;
}

static int parse_whitespace (const char *str, char **end);

#define UNIT_B (1UL)
#define UNIT_KiB (1024UL * UNIT_B)
#define UNIT_MiB (1024UL * UNIT_KiB)
#define UNIT_GiB (1024UL * UNIT_MiB)

static size_t
parse_unit (const char *str, char **end)
{
  char *my_end = (char *) str;
  char *test;
  if (parse_whitespace (my_end, &my_end))
    {
      return parse_unit (my_end, end);
    }
  test = _ ("GiB");
  if (strncmp (my_end, test, strlen (test)) == 0)
    {
      *end = my_end + strlen (test);
      return UNIT_GiB;
    }
  test = _ ("MiB");
  if (strncmp (my_end, test, strlen (test)) == 0)
    {
      *end = my_end + strlen (test);
      return UNIT_MiB;
    }
  test = _ ("KiB");
  if (strncmp (my_end, test, strlen (test)) == 0)
    {
      *end = my_end + strlen (test);
      return UNIT_KiB;
    }
  test = _ ("B");
  if (strncmp (my_end, test, strlen (test)) == 0)
    {
      *end = my_end + strlen (test);
      return UNIT_B;
    }
  *end = my_end;
  return UNIT_B;
}

static int
parse_whitespace (const char *str, char **end)
{
  if (strncmp (str, " ", strlen (" ")) == 0)
    {
      if (end)
	{
	  *end = ((char *) str) + strlen (" ");
	}
      return 1;
    }
  if (strncmp (str, " ", strlen (" ")) == 0)
    {
      if (end)
	{
	  *end = ((char *) str) + strlen (" ");
	}
      return 1;
    }
  return 0;
}
