/*
 * tarot_features.c
 *
 * Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "program_commons.h"
#include <tuneconfig.h>

static int encoded = 0;
static void parse_options (int argc, char *argv[]);

int
main (int argc, char *argv[])
{
  TarotGame *game = NULL;
  TarotGameIterator *it;
  const TarotGameEvent *event;
  TarotGame *rebuilt[5] = { NULL, NULL, NULL, NULL, NULL };
  int score[5];
  size_t n_players;
  size_t player;
  tarot_program_setup ();
  parse_options (argc, argv);
  game = tarot_program_read_game ();
  printf ("%s", tarot_features_arff_header (encoded));
  if (tarot_game_get_scores (game, &n_players, 0, 5, score) != TAROT_GAME_OK)
    {
      fprintf (stderr, _ ("Error: this game is not finished yet.\n"));
      return EXIT_FAILURE;
    }
  assert (n_players <= 5);
  it = tarot_game_iterator (game);
  for (player = 0; player < n_players; player++)
    {
      rebuilt[player] = tarot_game_alloc ();
    }
  while ((event = tarot_game_iterator_next_value (it)) != NULL)
    {
      TarotPlayer next;
      if (tarot_game_get_next (rebuilt[0], &next) == TAROT_GAME_OK)
	{
	  char *line = tarot_features_arff_data_alloc (rebuilt[next], event, score[next], encoded);
	  printf ("%s", line);
	  free (line);
	}
      switch (tarot_game_event_type (event))
	{
	case TAROT_DEAL_ALL_EVENT:
	  for (player = 0; player < n_players; player++)
	    {
	      TarotCard hand[78];
	      size_t n_hand;
	      TarotGameEvent *deal_local;
	      if (tarot_game_get_deal_of (game, player, &n_hand, 0, 78, hand) != TAROT_GAME_OK)
		{
		  char player_name[10];
		  tarot_player_to_string (player, sizeof (player_name), player_name);
		  player_name[sizeof (player_name) - 1] = '\0';
		  fprintf (stderr, _ ("Error: we don't know the cards of player %s.\n"),
			   player_name);
		  return EXIT_FAILURE;
		}
	      assert (n_hand < sizeof (hand) / sizeof (hand[0]));
	      deal_local = tarot_game_event_alloc_deal (player, n_hand, hand);
	      if (tarot_game_add_event (rebuilt[player], deal_local) != TAROT_GAME_OK)
		{
		  assert (0);
		}
	      tarot_game_event_free (deal_local);
	    }
	  break;
	case TAROT_DISCARD_EVENT:
	  if (1)
	    {
	      size_t n_full;
	      size_t n_part;
	      TarotCard full[78];
	      TarotCard part[78];
	      static const size_t max_full = sizeof (full) / sizeof (full[0]);
	      static const size_t max_part = sizeof (part) / sizeof (part[0]);
	      TarotPlayer taker;
	      TarotGameEvent *e_full;
	      TarotGameEvent *e_part;
	      tarot_game_get_taker (game, &taker);
	      if ((tarot_game_get_full_discard (game, &n_full, 0, max_full, full)
		   != TAROT_GAME_OK)
		  || (tarot_game_get_public_discard (game, &n_part, 0, max_part, part)
		      != TAROT_GAME_OK))
		{
		  fprintf (stderr, _ ("Error: we don't know the discard.\n"));
		  return EXIT_FAILURE;
		}
	      assert (n_full <= max_full);
	      assert (n_part <= max_part);
	      e_full = tarot_game_event_alloc_discard (n_full, full);
	      e_part = tarot_game_event_alloc_discard (n_part, part);
	      for (player = 0; player < n_players; player++)
		{
		  if (player == taker)
		    {
		      if (tarot_game_add_event (rebuilt[player], e_full) != TAROT_GAME_OK)
			{
			  assert (0);
			}
		    }
		  else
		    {
		      if (tarot_game_add_event (rebuilt[player], e_part) != TAROT_GAME_OK)
			{
			  assert (0);
			}
		    }
		}
	      tarot_game_event_free (e_full);
	      tarot_game_event_free (e_part);
	    }
	  break;
	default:
	  for (player = 0; player < n_players; player++)
	    {
	      if (tarot_game_add_event (rebuilt[player], event) != TAROT_GAME_OK)
		{
		  assert (0);
		}
	    }
	  break;
	}
    }
  return EXIT_SUCCESS;
}

static void
parse_options (int argc, char *argv[])
{
  int c;
  while (1)
    {
      static struct option long_options[] =
        {
	 {N_ ("help"), no_argument, 0, 'h'},
	 {N_ ("version"), no_argument, 0, 'v'},
	 {N_ ("encoded"), no_argument, 0, 'e'},
	 {0, 0, 0, 0}
        };
      int option_index = 0;
      for (option_index = 0; long_options[option_index].name != NULL; option_index++)
	{
	  long_options[option_index].name = _ (long_options[option_index].name);
	}
      option_index = 0;
      c = getopt_long (argc, argv, "hve", long_options, &option_index);
      if (c == -1)
        break;
      switch (c)
        {
        case 0:
          break;

	case 'h':
	  printf (_ ("Usage: tarot-features [OPTIONS]... < GAME > FEATURES\n\
Extract the game features.  Options:\n\
- -h, --help: print this message and exit;\n\
- -v, --version: print the program version and exit;\n\
- -e, --encoded: one-hot encode the categorical features.\n\
"));
	  exit (EXIT_SUCCESS);
	  break;

	case 'v':
	  printf (_ ("%s (libtarot %s)\n\
\n\
Copyright © 2019 Vivien Kraus\nThis is free software; see the source for copying conditions.  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
		  "tarot-features",
		  tarot_libtarot_version ());
	  exit (EXIT_SUCCESS);
	  break;

	case 'e':
	  encoded = 1;
	  break;

        case '?':
          /* getopt_long already printed an error message. */
	  exit (EXIT_FAILURE);
          break;

        default:
          abort ();
        }
    }
  if (optind < argc)
    {
      fprintf (stderr, _ ("Error: extra arguments:\n"));
      while (optind < argc)
        fprintf (stderr, _ ("- %s\n"), argv[optind++]);
      exit (EXIT_FAILURE);
    }
}
